import shutil
import os

assets_directory = 'tests/assets'


def pytest_sessionstart(session):
    """ before session.main() is called. """
    if not os.path.exists(assets_directory):
        os.makedirs(assets_directory)


def pytest_sessionfinish(session, exitstatus):
    """ whole test run finishes. """
    shutil.rmtree(assets_directory)


from mira import detectors, core


def test_input_output():
    annotation_config = core.AnnotationConfiguration(['foo', 'bar'])
    collection = core.SceneCollection(
        annotation_config=annotation_config,
        scenes=[
            core.Scene(
                image=core.Image.new(
                    width=100,
                    height=100,
                    channels=3
                ),
                annotation_config=annotation_config,
                annotations=[
                    core.Annotation(
                        category=annotation_config['foo'],
                        selection=core.Selection([[0, 0], [50, 50]])
                    )
                ]
            ),
            core.Scene(
                image=core.Image.new(
                    width=100,
                    height=100,
                    channels=3
                ),
                annotation_config=annotation_config,
                annotations=[
                    core.Annotation(
                        category=annotation_config['bar'],
                        selection=core.Selection([[20, 20], [80, 80]])
                    )
                ]
            ),
        ]
    )

    rn = detectors.RetinaNet(
        pretrained_backbone=False,
        annotation_config=annotation_config
    )
    yolo = detectors.YOLOv3(
        pretrained_backbone=False,
        annotation_config=annotation_config
    )
    east = detectors.EAST(
        pretrained_backbone=False,
        annotation_config=annotation_config,
        text_category='foo'
    )
    advancedeast = detectors.AdvancedEAST(
        pretrained_backbone=False,
        annotation_config=annotation_config,
        text_category='foo'
    )

    inverted = rn.invert_targets(y=rn.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[1].annotations[0].selection.bbox(), [20, 20, 80, 80]))  # noqa: E501

    inverted = yolo.invert_targets(y=yolo.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[1].annotations[0].selection.bbox(), [20, 20, 80, 80]))  # noqa: E501

    inverted = east.invert_targets(y=east.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 1 <= a < e + 1 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert len(inverted[1].annotations) == 0

    inverted = advancedeast.invert_targets(y=advancedeast.compute_targets(collection), images=collection.images)  # noqa: E501
    assert all(e - 2 <= a < e + 2 for a, e in zip(inverted[0].annotations[0].selection.bbox(), [0, 0, 50, 50]))  # noqa: E501
    assert len(inverted[1].annotations) == 0



from typing import List


def compute_contained(si: List[int], so: List[int]):
    """Compute the percentage of one box that is contained
    within another box.

    Args:
        si: A list of four values, xi, yi, wi, yi
            for the inner box
        so: A list of four values, xo, yo, wo, yo
            for the outer box

    Returns:
        The total overlap
    """
    xi, yi, wi, hi = si
    xo, yo, wo, ho = so

    # Find the width and height of the box within the inner
    # box that is missing from the outer box
    wnc = max(min(xo-xi, wi), 0) + max(min((xi + wi) - (xo+wo), wi), 0)
    hnc = max(min(yo-yi, hi), 0) + max(min((yi + hi) - (yo+ho), hi), 0)

    # Compute the size of the contained box
    wc = wi - wnc
    hc = hi - hnc

    inner = (hi*wi)
    if inner == 0:
        return 0
    return (wc*hc) / inner



from mira import datasets, core

test_string = """
<annotation>
    <folder>VOC2012</folder>
    <filename>test_voc.jpg</filename>
    <source>
        <database>The VOC2007 Database</database>
        <annotation>PASCAL VOC2007</annotation>
        <image>flickr</image>
    </source>
    <size>
        <width>486</width>
        <height>500</height>
        <depth>3</depth>
    </size>
    <segmented>0</segmented>
    <object>
        <name>person</name>
        <pose>Unspecified</pose>
        <truncated>0</truncated>
        <difficult>0</difficult>
        <bndbox>
            <xmin>174</xmin>
            <ymin>101</ymin>
            <xmax>349</xmax>
            <ymax>351</ymax>
        </bndbox>
        <part>
            <name>head</name>
            <bndbox>
                <xmin>169</xmin>
                <ymin>104</ymin>
                <xmax>209</xmax>
                <ymax>146</ymax>
            </bndbox>
        </part>
        <part>
            <name>hand</name>
            <bndbox>
                <xmin>278</xmin>
                <ymin>210</ymin>
                <xmax>297</xmax>
                <ymax>233</ymax>
            </bndbox>
        </part>
        <part>
            <name>foot</name>
            <bndbox>
                <xmin>273</xmin>
                <ymin>333</ymin>
                <xmax>297</xmax>
                <ymax>354</ymax>
            </bndbox>
        </part>
        <part>
            <name>foot</name>
            <bndbox>
                <xmin>319</xmin>
                <ymin>307</ymin>
                <xmax>340</xmax>
                <ymax>326</ymax>
            </bndbox>
        </part>
    </object>
</annotation>
"""


def test_voc():
    with open('tests/assets/test_voc.xml', 'w') as f:
        f.write(test_string)
    core.Image.new(
        width=486,
        height=500,
        channels=3
    ).save(
        'tests/assets/test_voc.jpg'
    )

    original = datasets.load_voc(
        filepaths=['tests/assets/test_voc.xml'],
        annotation_config=core.AnnotationConfiguration(['person']),
        image_dir='tests/assets/'
    )
    assert len(original[0].annotations) == 1


from mira.core import Image


def test_blank_and_properties():
    image = Image.new(width=200, height=100, channels=2, cval=125)
    assert image.height == 100
    assert image.width == 200
    assert image.channels == 2
    assert (image[:, :, :] == 125).all()


def test_file_read():
    image = Image.new(width=200, height=100, channels=3, cval=125)
    image[40:60, 40:60, 0] = 0
    image.save('tests/assets/test.jpg')
    image = Image.read('tests/assets/test.jpg')

    def consistency_check(img):
        assert img.width == 200
        assert img.height == 100
        assert img.channels == 3
        assert (img[45:55, 45:55, 0] <= 5).all()
        assert (img[45:55, 45:55, 1:] > 100).all()

    consistency_check(image)
    with open('tests/assets/test.jpg', 'rb') as buffer:
        image = Image.read(buffer)
    consistency_check(image)

    image = Image.read(image.buffer('.jpeg'))
    consistency_check(image)


import csv
import json
from os import path
from typing import Union

import validators

from ..core import (
    Annotation,
    Scene,
    SceneCollection,
    Selection,
    AnnotationConfiguration
)


def load_via(
    annotations_file: str,
    label_key: str='class',
    image_dir: str=None,
    annotation_config: AnnotationConfiguration=None,
    cache: Union[str, bool]=True
) -> SceneCollection:
    """Load annotations created by the `VGG Image Annotation Tool
    <http://www.robots.ox.ac.uk/~vgg/software/via/>`_.

    Args:
        annotations_file: The file containing the
            annotations. For now, only the CSV
            format is supported.
        label_key: The attribute name to use for class labels
        annotation_config: The annotation configuration to use.
            If None, it is inferred from the annotations file.
        image_dir: The directory in which to look
            for images. If None, defaults the
            folder in which the annotations_file
            is stored.
        cache: The cache parameter to pass to `Scene`. Useful
            if the annotations file references URLs (i.e.,
            specify a directory in which to save images).
    """
    if not annotations_file.endswith('.csv'):
        raise NotImplementedError(
            'Only the VIA CSV format is supported at '
            'this time.'
        )
    if image_dir is None:
        image_dir = path.split(annotations_file)[0]
    annotations_dict = {}
    category_set = set()
    with open(annotations_file) as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='"')
        for i, row in enumerate(reader):
            if i == 0:
                continue
            image_name = row[0]
            if image_name not in annotations_dict:
                annotations_dict[image_name] = []

            region_shape = json.loads(row[5])
            region_attr = json.loads(row[6])

            if len(region_shape) == 0:
                # If there is no region information,
                # it means there is no annotation in this
                # image.
                continue

            if region_shape['name'] == 'rect':
                x = region_shape['x'],
                y = region_shape['y'],
                width = region_shape['width'],
                height = region_shape['height']
                selection = Selection(
                    [
                        [x, y],
                        [x + width, y + height]
                    ]
                )
            else:
                raise NotImplementedError(
                    'Only rectangles are supported at this time, '
                    'not %s' % region_shape['name']
                )

            category_name = region_attr[label_key].strip()
            category_set.update([category_name])
            annotations_dict[image_name].append(
                (category_name, selection)
            )
    if annotation_config is not None:
        assert all(c in annotation_config for c in category_set)
    if annotation_config is None:
        annotation_config = AnnotationConfiguration(
            sorted(list(category_set))
        )
    scenes = []
    for image, annotations in annotations_dict.items():
        if not validators.url(image):
            image = path.join(image_dir, image)
        annotations = [
            Annotation(
                selection=selection,
                category=annotation_config[category_name]
            ) for category_name, selection in annotations
        ]
        scenes.append(
            Scene(
                image=image,
                annotations=annotations,
                annotation_config=annotation_config,
                cache=cache
            )
        )
    return SceneCollection(
        scenes=scenes,
        annotation_config=annotation_config
    )